import abc
from typing import Optional, Union

from unfolded.map_sdk.api.dataset_api import (
    DatasetApiInteractiveMixin,
    DatasetApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.event_api import EventApiInteractiveMixin
from unfolded.map_sdk.api.filter_api import (
    FilterApiInteractiveMixin,
    FilterApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.layer_api import (
    LayerApiInteractiveMixin,
    LayerApiNonInteractiveMixin,
)
from unfolded.map_sdk.api.map_api import (
    MapApiInteractiveMixin,
    MapApiNonInteractiveMixin,
)
from unfolded.map_sdk.transport.base import (
    BaseInteractiveTransport,
    BaseNonInteractiveTransport,
    BaseTransport,
)


class BaseMap:
    """
    Base class for all map types (both widget and non-widget)
    """

    transport: BaseTransport

    @abc.abstractmethod
    def __init__(
        self,
        *,
        map_url: Optional[str] = None,
        map_uuid: Optional[str] = None,
        width: Optional[Union[str, int, float]] = None,
        height: Optional[Union[str, int, float]] = None,
    ):
        pass


class BaseInteractiveMap(
    BaseMap,
    MapApiInteractiveMixin,
    DatasetApiInteractiveMixin,
    FilterApiInteractiveMixin,
    LayerApiInteractiveMixin,
    EventApiInteractiveMixin,
):
    transport: BaseInteractiveTransport
    pass


class BaseNonInteractiveMap(
    BaseMap,
    MapApiNonInteractiveMixin,
    DatasetApiNonInteractiveMixin,
    FilterApiNonInteractiveMixin,
    LayerApiNonInteractiveMixin,
):
    transport: BaseNonInteractiveTransport
    pass



from typing import Dict, List, Optional

from ipywidgets import DOMWidget
from pydantic import AnyHttpUrl
from traitlets import Dict as TraitletsDict
from traitlets import Unicode

from unfolded.map_sdk._frontend import module_name, module_version
from unfolded.map_sdk._version import __version__
from unfolded.map_sdk.api.base import CamelCaseBaseModel
from unfolded.map_sdk.api.map_api import MapStyleCreationProps
from unfolded.map_sdk.environment import default_height
from unfolded.map_sdk.map.base import BaseInteractiveMap
from unfolded.map_sdk.transport.widget import BlockingWidgetTransport


class BasemapParams(CamelCaseBaseModel):
    custom_map_styles: Optional[List[MapStyleCreationProps]]
    initial_map_style_id: Optional[str]


class RasterParams(CamelCaseBaseModel):
    server_urls: Optional[List[AnyHttpUrl]]


DEFAULT_MAP_STYLE = {
    "height": default_height(),
    "width": "100%",
}


class SyncWidgetMap(DOMWidget, BaseInteractiveMap):
    _model_name = Unicode("UnfoldedMapModel").tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode("UnfoldedMapView").tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    style = TraitletsDict().tag(sync=True)
    basemaps = TraitletsDict().tag(sync=True)
    raster = TraitletsDict().tag(sync=True)
    _internal = TraitletsDict().tag(sync=True)

    # Note: We define an unused kwargs so that an error is not produced for invalid initialization
    # args between the widget and HTML implementations
    def __init__(
        self,
        *,
        style: Optional[Dict] = None,
        basemaps: Optional[Dict] = None,
        raster: Optional[Dict] = None,
        _internal: Optional[Dict] = None,
        **kwargs  # pylint: disable=unused-argument
    ):
        """Initializes a new widget map

        Kwargs:
            style: Optional map container CSS style customization. Uses camelCase as this is React standard.
            basemaps: Basemap customization settings.
            raster: Customization related to raster datasets and tiles.

        """
        super().__init__()
        self.transport = BlockingWidgetTransport(widget=self)

        if style:
            self.style = {**DEFAULT_MAP_STYLE, **style}
        else:
            self.style = DEFAULT_MAP_STYLE

        if basemaps:
            validated_basemaps = BasemapParams(**basemaps)
            self.basemaps = validated_basemaps.dict(by_alias=True, exclude_none=True)

        if raster:
            validated_raster = RasterParams(**raster)
            self.raster = validated_raster.dict(by_alias=True, exclude_none=True)

        if _internal:
            self._internal = _internal

        on_msg = lambda widget, content, buffers: self.transport.receive_message(
            content, buffers
        )
        self.on_msg(on_msg)
        self._is_loaded = False

    def render(self) -> None:
        raise NotImplementedError()



from unfolded.map_sdk.api import filter_api

DUMMY_MESSAGE_UUID = "ede3fafc-4945-41b7-ac64-16faa896a47a"
DATASET_UUID = "0e939627-8ea7-4db6-8c92-3dd943166a01"
LAYER_UUID = "997c21eb-604b-49df-add9-b2fcc615a243"
FILTER_UUID = "e02b67ea-20d4-4613-808e-9a8fdbd81e6f"

ACTION_FIXTURES = [
    (
        filter_api.GetFiltersAction(),
        {
            "type": "v1/map-sdk-get-filters",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [],
        },
    ),
    (
        filter_api.GetFilterByIdAction(filter_id=FILTER_UUID),
        {
            "type": "v1/map-sdk-get-filter-by-id",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [FILTER_UUID],
        },
    ),
    (
        filter_api.AddFilterAction(
            filter={
                "type": filter_api.FilterType.RANGE,
                "sources": [{"data_id": DATASET_UUID, "field_name": "test"}],
                "value": (0, 100),
            }
        ),
        {
            "type": "v1/map-sdk-add-filter",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                {
                    "type": "range",
                    "sources": [{"dataId": DATASET_UUID, "fieldName": "test"}],
                    "value": [0, 100],
                }
            ],
        },
    ),
    (
        filter_api.UpdateFilterAction(
            filter_id=FILTER_UUID,
            values={
                "value": (0, 50),
                "sources": [{"data_id": DATASET_UUID, "field_name": "test-2"}],
            },
        ),
        {
            "type": "v1/map-sdk-update-filter",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [
                FILTER_UUID,
                {
                    "type": "range",
                    "value": [0, 50],
                    "sources": [{"dataId": DATASET_UUID, "fieldName": "test-2"}],
                },
            ],
        },
    ),
    (
        filter_api.RemoveFilterAction(filter_id=FILTER_UUID),
        {
            "type": "v1/map-sdk-remove-filter",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [FILTER_UUID],
        },
    ),
    (
        filter_api.UpdateTimelineAction(
            filter_id=FILTER_UUID, values={"view": "side", "is_animating": True}
        ),
        {
            "type": "v1/map-sdk-update-timeline",
            "messageId": DUMMY_MESSAGE_UUID,
            "args": [FILTER_UUID, {"view": "side", "isAnimating": True}],
        },
    ),
]



import logging
from typing import Dict, List, Optional, Tuple

from unfolded.map_sdk.api.base import Action
from unfolded.map_sdk.utils.encoders import jsonable_encoder


def serialize_action(action: Action) -> Tuple[Dict, Optional[List[bytes]]]:
    """Serialize Action to format that can be sent through Jupyter Comm mechanism"""
    # Use FastAPI's encoder to serialize to dict directly, without going through json
    # https://fastapi.tiangolo.com/tutorial/encoder/
    # https://stackoverflow.com/a/68778590
    d = jsonable_encoder(action, by_alias=True, exclude_none=True)

    # We want all keys except for `type` and `messageId` to be within a top-level `data` key

    # Make sure the args key doesn't exist yet
    if d.get("args") != None:
        logging.debug("args key already exists")

    # We define new objects instead of assigning to `d` immediately so that an attribute of the
    # action can be named `args` or `options`
    new_args = []
    new_options = {}

    # Keys that should stay at the top level
    top_level_json_keys = ["type", "messageId"]

    options_keys = list(map(action.Config.alias_generator, action.Meta.options))

    # Get all non-top-level non-options argument keys
    model_keys = set(d.keys()).difference(top_level_json_keys).difference(options_keys)

    # Transform listed args to (usually CamelCase) aliases
    arg_keys = list(map(action.Config.alias_generator, action.Meta.args))

    if model_keys != set(arg_keys):
        mismatched_keys = model_keys.symmetric_difference(arg_keys)
        logging.debug(
            f"Mismatch between args list and model fields in model {action.__class__.__name__}: {mismatched_keys}"
        )

    # Add arguments to args list
    for key in arg_keys:
        new_args.append(d.pop(key))

    # Add options to options dict
    for key in options_keys:
        val = d.pop(key)
        if val is not None:
            new_options[key] = val

    d["args"] = new_args

    # Don't send options if empty
    if new_options:
        d["options"] = new_options

    # Not currently using binary Comm message transfer
    return d, None



import abc
from typing import Callable, Dict, List, Optional, Sequence, Type, Union

from unfolded.map_sdk.api.base import Action
from unfolded.map_sdk.api.enums import EventType
from unfolded.map_sdk.types import (
    ErrorResponse,
    EventResponse,
    MessageResponse,
    ResponseClass,
)


class BaseTransport(abc.ABC):
    @abc.abstractmethod
    def send_action(
        self, *, action: Action, response_class: Optional[Type[ResponseClass]] = None
    ) -> Optional[ResponseClass]:
        pass


class BaseInteractiveTransport(BaseTransport):
    # Note: this is stored as a dict instead of the EventHandlers model to avoid circular imports
    event_handlers: Dict[str, Callable]

    @abc.abstractmethod
    def send_action(
        self, *, action: Action, response_class: Optional[Type[ResponseClass]] = None
    ) -> Optional[ResponseClass]:
        pass

    @abc.abstractmethod
    def send_action_non_null(
        self,
        *,
        action: Action,
        response_class: Optional[Type[ResponseClass]] = None,
    ) -> ResponseClass:
        pass

    @abc.abstractmethod
    def receive_message(
        self,
        content: Union[EventResponse, MessageResponse, ErrorResponse],
        buffers: List[bytes],
    ) -> None:
        pass

    @abc.abstractmethod
    def set_event_handlers(self, event_handlers: Dict[str, Callable]) -> None:
        pass

    @abc.abstractmethod
    def remove_event_handlers(self, names: Sequence[Union[str, EventType]]) -> None:
        pass


class BaseNonInteractiveTransport(BaseTransport):
    @abc.abstractmethod
    def send_action(
        self, *, action: Action, response_class: Optional[Type[ResponseClass]] = None
    ) -> None:
        pass



from typing import Any, Optional

from unfolded.map_sdk.environment import CURRENT_ENVIRONMENT, Environment
from unfolded.map_sdk.map import SyncWidgetMap

__all__ = ("create_map",)


def create_map(*, html: Optional[bool] = None, **kwargs: Any) -> SyncWidgetMap:
    """Create an unfolded map

    Args:
        html (Optional[bool], optional): Whether to create a pure HTML map. Defaults to None.
    """
    if html or CURRENT_ENVIRONMENT == Environment.DATABRICKS:
        raise NotImplementedError("Static HTML map not yet implemented")
        # return HTMLMap(**kwargs)

    return SyncWidgetMap(**kwargs)



import sys


def enable_colab_custom_widgets() -> None:
    """Enable custom widget support in Colab"""
    # Uses same approach as ipyleaflet
    # https://github.com/jupyter-widgets/ipyleaflet/pull/871
    if "google.colab" in sys.modules:
        from google.colab import output  # pylint:disable=import-outside-toplevel

        output.enable_custom_widget_manager()



from pathlib import Path
from typing import Optional, Union

from unfolded.map_sdk.map.base import BaseNonInteractiveMap
from unfolded.map_sdk.transport.html import HTMLTransport

TEMPLATE_DIR = (Path(__file__).parent / ".." / "templates").resolve()


class HTMLMap(BaseNonInteractiveMap):
    transport: HTMLTransport

    map_url: Optional[str]
    map_uuid: Optional[str]
    width: Optional[Union[str, int, float]]
    height: Optional[Union[str, int, float]]

    def __init__(
        self,
        *,
        map_url: Optional[str] = None,
        map_uuid: Optional[str] = None,
        width: Optional[Union[str, int, float]] = None,
        height: Optional[Union[str, int, float]] = None,
    ):
        super().__init__()
        self.map_url = map_url
        self.map_uuid = map_uuid
        self.width = width
        self.height = height
        self.transport = HTMLTransport()
        self.rendered = False

    def _repr_html_(self) -> str:
        return self.render()

    def render(self) -> str:
        return self.transport.render_template()



from functools import wraps
from typing import Any, Callable, Optional, Sequence, TypeVar

from unfolded.map_sdk.errors import MapSDKException

__all__ = ("validate_kwargs", "assert_not_rendered")

FuncT = TypeVar("FuncT", bound=Callable[..., Any])


def validate_kwargs(n_args: int = 2, positional_only: Optional[Sequence[str]] = None):
    """Checks to make sure both positional args and kwargs are not used"""
    # Note the default for n_args is 2 to account for self

    def decorator(func: FuncT) -> FuncT:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if len(args) >= n_args and len(kwargs) > 0:
                raise MapSDKException(
                    "Cannot pass in both positional arguments and kwargs, use one or the other."
                )

            # Language support for declaring positional-only arguments with / began in 3.8, and we
            # support 3.7
            # https://peps.python.org/pep-0570/
            if positional_only:
                positional_only_used = set(kwargs.keys()).intersection(positional_only)
                if len(positional_only_used) == 1:
                    raise MapSDKException(
                        f"'{list(positional_only_used)[0]}' is a positional-only argument."
                    )
                elif len(positional_only_used) > 1:
                    arg_list = ", ".join([f"'{arg}'" for arg in positional_only_used])
                    raise MapSDKException(f"{arg_list} are positional-only arguments.")

            return func(*args, **kwargs)

        # Incompatible return value type (got "Callable[[VarArg(Any), KwArg(Any)], Any]", expected "FuncT")
        return wrapper  # type:ignore[return-value]

    return decorator


def assert_not_rendered(func):
    """Assert that the class has not yet been rendered before calling this function"""

    @wraps(func)
    def wrap(self, *args, **kwargs):
        if self.rendered:
            raise MapSDKException(
                f"Cannot call {func.__name__} on already-rendered map"
            )

        return func(self, *args, **kwargs)

    return wrap
